//--------------------------------------------------------------------------------------
// File: ShadowDX9.fx
//
// The effect file for the ShadowDX9 sample.  
// 
// Copyright (c) Microsoft Corporation. All rights reserved.
//--------------------------------------------------------------------------------------


//--------------------------------------------------------------------------------------
// Global variables
//--------------------------------------------------------------------------------------
float4 g_MaterialAmbientColor;      // Material's ambient color
float4 g_MaterialDiffuseColor;      // Material's diffuse color
int g_nNumLights;

float3 g_LightDir[3];               // Light's direction in world space
float4 g_LightDiffuse[3];           // Light's diffuse color
float4 g_LightAmbient;              // Light's ambient color

texture g_MeshTexture;              // Color texture for mesh

float    g_fTime;                   // App's time in seconds
float4x4 g_mWorld;                  // World matrix for object
float4x4 g_mWorldViewProjection;    // World * View * Projection matrix
float4x4 g_mLocalToShadowbufferProjection;	// Shadowbuffer projection matrix.

float	g_fShadowbufferRefValue;	// The value the SB read is compared against
float   g_fShadowbufferTexelSpread;	// how far apart to spread the texel samples in UV space.

float   g_fShadowbufferDepthDebugScale;

float	g_fShadowbufferDepthOffset;


//--------------------------------------------------------------------------------------
// Texture samplers
//--------------------------------------------------------------------------------------
sampler ShadowBufferSampler = 
sampler_state
{
    Texture = <g_MeshTexture>;
    MipFilter = LINEAR;
    MinFilter = LINEAR;
    MagFilter = LINEAR;
    AddressU = CLAMP;
    AddressV = CLAMP;
};

sampler ShadowBufferSamplerPoint = 
sampler_state
{
    Texture = <g_MeshTexture>;
    MipFilter = POINT;
    MinFilter = POINT;
    MagFilter = POINT;
    AddressU = CLAMP;
    AddressV = CLAMP;
};


//--------------------------------------------------------------------------------------
// Vertex shader output structure
//--------------------------------------------------------------------------------------
struct VS_OUTPUT
{
    float4 Position   : POSITION;   // vertex position 
    float4 Diffuse    : COLOR0;     // vertex diffuse color (note that COLOR0 is clamped from 0..1)
    float3 TextureUVW : TEXCOORD0;  // vertex texture coords (2D with projection)
    float2 SbDepth	  : TEXCOORD1;  // depth of surface at this point in the shadowbuffer
};


//--------------------------------------------------------------------------------------
// This shader computes standard transform and lighting
//--------------------------------------------------------------------------------------
VS_OUTPUT RenderSceneVS(		float3 vPos : POSITION, 
								float3 vNormal : NORMAL,
								uniform bool bLighting,
								uniform bool bProjectionTexture )
{
    VS_OUTPUT Output;
    float3 vNormalWorldSpace;
  
    float4 vFullPos = float4 ( vPos, 1.0f );
      
    // Transform the position from object space to homogeneous projection space
    Output.Position = mul(vFullPos, g_mWorldViewProjection);
    
    // Transform the normal from object space to world space    
    vNormalWorldSpace = normalize(mul(vNormal, (float3x3)g_mWorld)); // normal (world space)
    
    if ( bLighting )
    {
		// Compute simple directional lighting equation
		float3 vTotalLightDiffuse = float3(0,0,0);
		const int nNumLights = 1;
		for(int i=0; i<nNumLights; i++ )
		{
			vTotalLightDiffuse += g_LightDiffuse[i] * max(0,dot(vNormalWorldSpace, g_LightDir[i]));
		}

		// Note - no ambient
		Output.Diffuse.rgb = g_MaterialDiffuseColor * vTotalLightDiffuse;
		Output.Diffuse.a = 1.0f; 
	}
	else
	{
		// Just ambient
		Output.Diffuse.rgb = g_MaterialAmbientColor * g_LightAmbient;
		Output.Diffuse.a = 1.0f; 
	}
	
	float4 ShadowbufferPosition = mul(vFullPos, g_mLocalToShadowbufferProjection);
	if ( bProjectionTexture )
	{
		Output.TextureUVW = float3 ( ShadowbufferPosition.x, ShadowbufferPosition.y, ShadowbufferPosition.w );
	}
	else
	{
		Output.TextureUVW = float3 ( 0.0f, 0.0f, 0.0f );
	}
	
	// Make sure the pixel shader gets to see the SB depth.
	//Output.SbDepth = ( ShadowbufferPosition.z / ShadowbufferPosition.w );
	Output.SbDepth.x = ShadowbufferPosition.z;
	Output.SbDepth.y = ShadowbufferPosition.w;
	
    return Output;
}

//--------------------------------------------------------------------------------------
// Pixel shader output structure
//--------------------------------------------------------------------------------------
struct PS_OUTPUT
{
    float4 RGBColor : COLOR0;  // Pixel color
};


PS_OUTPUT RenderScenePS (	VS_OUTPUT In,
							uniform bool bShadowbufferProject,
							uniform bool bShadowbufferRedDepth,
							uniform bool bShadowbufferIdCompare,
							uniform bool bShadowbufferDepthCompare,
							uniform bool bShadowbufferDepthCompare16,
							uniform bool bShadowbufferDebugShow,
							uniform bool bShadowbufferDebugRed,
							uniform bool bShadowbufferDebugAlpha ) 
{ 
    PS_OUTPUT Output;

   	Output.RGBColor.rgb = In.Diffuse;
	Output.RGBColor.a = g_fShadowbufferRefValue;
	
	float SbDepth = In.SbDepth.x / In.SbDepth.y;
	if ( bShadowbufferRedDepth )
	{
		// Red channel is the 8-bit depth
		Output.RGBColor.r = SbDepth;
		
		// Green+blue channels are the 16-bit depth
		float Hi, Lo;
		Lo = modf ( SbDepth * 256, Hi );
		Output.RGBColor.g = Hi / 256;
		Output.RGBColor.b = Lo;
	}

    if ( bShadowbufferProject )
    {
		float OneOverW = 1.0f / In.TextureUVW.z;
		float2 TexCoord = float2 ( In.TextureUVW.x * OneOverW, In.TextureUVW.y * OneOverW );
		
		if ( bShadowbufferDebugShow )
		{
			// Debug - show the value.
			float4 ShadowbufferValue = tex2D ( ShadowBufferSamplerPoint, TexCoord );
			
			float DepthValue = ShadowbufferValue.r;		
			if ( bShadowbufferDepthCompare16 )
			{
				// Reassemble a depth value from the g and b channels.
				DepthValue = ShadowbufferValue.g + ( ShadowbufferValue.b / 256 );
			}
			float Ignored;
			DepthValue = modf ( DepthValue * g_fShadowbufferDepthDebugScale, Ignored );
			
			if ( bShadowbufferDebugAlpha )
			{
				Output.RGBColor.rgb = ShadowbufferValue.a;
			}
			else if ( bShadowbufferDebugRed )
			{
				Output.RGBColor.rgb = DepthValue;
			}
			else
			{
				Output.RGBColor.rgb = ShadowbufferValue.rgb;
			}
		}
		else
		{
			if ( bShadowbufferIdCompare )
			{
				float4 ShadowbufferValue00 = tex2D ( ShadowBufferSamplerPoint, TexCoord + float2 ( -g_fShadowbufferTexelSpread, -g_fShadowbufferTexelSpread ) );
				float4 ShadowbufferValue01 = tex2D ( ShadowBufferSamplerPoint, TexCoord + float2 ( -g_fShadowbufferTexelSpread,  g_fShadowbufferTexelSpread ) );
				float4 ShadowbufferValue10 = tex2D ( ShadowBufferSamplerPoint, TexCoord + float2 (  g_fShadowbufferTexelSpread, -g_fShadowbufferTexelSpread ) );
				float4 ShadowbufferValue11 = tex2D ( ShadowBufferSamplerPoint, TexCoord + float2 (  g_fShadowbufferTexelSpread,  g_fShadowbufferTexelSpread ) );

				// Discard pixel if all of ( ShadowbufferValueXX.w != g_fShadowbufferRefValue )
				//float Difference = abs ( ShadowbufferValue.w - g_fShadowbufferRefValue );
				float Difference;
				Difference  = ( ShadowbufferValue00.w - g_fShadowbufferRefValue );
				Difference *= ( ShadowbufferValue01.w - g_fShadowbufferRefValue );
				Difference *= ( ShadowbufferValue10.w - g_fShadowbufferRefValue );
				Difference *= ( ShadowbufferValue11.w - g_fShadowbufferRefValue );
				// For an 8-bit ID buffer, the smallest delta is 1/256, so if all four are that, the number will be 1/(2^32), or about 1e-10
				clip ( 1e-10f - abs ( Difference ) );

				// Lit. Or at least, until the depth check.
			}
			
			if ( bShadowbufferDepthCompare )
			{
				float4 ShadowbufferValue = tex2D ( ShadowBufferSamplerPoint, TexCoord );
				
				// If we're not doing ID buffering as well, do the depth comparison
				// If we are, only do the depth comparison if this is the same object.
				if ( !bShadowbufferIdCompare ||
					  ( abs ( ShadowbufferValue.w - g_fShadowbufferRefValue ) < 0.001 ) )
				{
					float DepthValue = ShadowbufferValue.r;
					if ( bShadowbufferDepthCompare16 )
					{
						// Reassemble a depth value from the g and b channels.
						DepthValue = ShadowbufferValue.g + ( ShadowbufferValue.b / 256 );
					}
					
	#if 0
	// Debug stuff.
					float2 DisplayedDepthValue = float2 ( SbDepth, DepthValue );
					float2 Ignored;
					DisplayedDepthValue = modf ( DisplayedDepthValue * g_fShadowbufferDepthDebugScale, Ignored );
					Output.RGBColor.rg = DisplayedDepthValue;
					Output.RGBColor.b = 0;
					
					if ( SbDepth > DepthValue + g_fShadowbufferDepthOffset )
					{
						Output.RGBColor.b = 1;
					}
					else
					{
						Output.RGBColor.b = 0;
					}
	#else
					clip ( DepthValue - SbDepth + g_fShadowbufferDepthOffset );
	#endif
				}
			}
		}
    }

    return Output;
}



//--------------------------------------------------------------------------------------
// Renders scene to render target
//--------------------------------------------------------------------------------------
technique RenderSceneWithAmbient
{
    pass P0
    {          
		ALPHABLENDENABLE = false;
		ALPHATESTENABLE = false;
        VertexShader = compile vs_2_0 RenderSceneVS( false, false );
        PixelShader  = compile ps_2_0 RenderScenePS( false, false, false, false, false, false, false, false );
    }
}

technique RenderSceneWithLightNoShadow
{
    pass P0
    {
		ALPHABLENDENABLE = true;
		SRCBLEND = ONE;
		DESTBLEND = ONE;
		ALPHATESTENABLE = false;
        VertexShader = compile vs_2_0 RenderSceneVS( true, false );
        PixelShader  = compile ps_2_0 RenderScenePS( false, false, false, false, false, false, false, false );
    }
}

technique RenderSceneWithLightWithShadowId
{
    pass P0
    {
		ALPHABLENDENABLE = true;
		SRCBLEND = ONE;
		DESTBLEND = ONE;
		ALPHATESTENABLE = false;
        VertexShader = compile vs_2_0 RenderSceneVS( true, true );
        PixelShader  = compile ps_2_0 RenderScenePS( true, false, true, false, false, false, false, false );
    }
}

technique RenderSceneWithLightWithShadowIdLocalDepth
{
    pass P0
    {
		ALPHABLENDENABLE = true;
		SRCBLEND = ONE;
		DESTBLEND = ONE;
		ALPHATESTENABLE = false;
        VertexShader = compile vs_2_0 RenderSceneVS( true, true );
        PixelShader  = compile ps_2_0 RenderScenePS( true, false, true, true, false, false, false, false );
    }
}

technique RenderSceneWithLightWithShadowDepth
{
    pass P0
    {
		ALPHABLENDENABLE = true;
		SRCBLEND = ONE;
		DESTBLEND = ONE;
		ALPHATESTENABLE = false;
        VertexShader = compile vs_2_0 RenderSceneVS( true, true );
        PixelShader  = compile ps_2_0 RenderScenePS( true, false, false, true, true, false, false, false );
    }
}

technique RenderSceneWithShadowbufferColour
{
    pass P0
    {
		ALPHABLENDENABLE = false;
		ALPHATESTENABLE = false;
        VertexShader = compile vs_2_0 RenderSceneVS( true, true );
        PixelShader  = compile ps_2_0 RenderScenePS( true, false, false, false, false, true, false, false );
    }
}

technique RenderSceneWithShadowbufferAlpha
{
    pass P0
    {
		ALPHABLENDENABLE = false;
		ALPHATESTENABLE = false;
        VertexShader = compile vs_2_0 RenderSceneVS( true, true );
        PixelShader  = compile ps_2_0 RenderScenePS( true, false, false, false, false, true, false, true );
    }
}

technique RenderSceneWithShadowbufferDepth16
{
    pass P0
    {
		ALPHABLENDENABLE = false;
		ALPHATESTENABLE = false;
        VertexShader = compile vs_2_0 RenderSceneVS( true, true );
        PixelShader  = compile ps_2_0 RenderScenePS( true, false, false, false, true, true, true, false );
    }
}

technique RenderSceneWithShadowbufferDepth8
{
    pass P0
    {
		ALPHABLENDENABLE = false;
		ALPHATESTENABLE = false;
        VertexShader = compile vs_2_0 RenderSceneVS( true, true );
        PixelShader  = compile ps_2_0 RenderScenePS( true, false, false, false, false, true, true, false );
    }
}

technique RenderSceneToShadowbufferId
{
    pass P0
    {          
		ALPHABLENDENABLE = false;
		ALPHATESTENABLE = false;
        VertexShader = compile vs_2_0 RenderSceneVS( true, false );
        PixelShader  = compile ps_2_0 RenderScenePS( false, false, true, false, false, false, false, false );
    }
}

technique RenderSceneToShadowbufferIdDepth
{
    pass P0
    {          
		ALPHABLENDENABLE = false;
		ALPHATESTENABLE = false;
        VertexShader = compile vs_2_0 RenderSceneVS( true, false );
        PixelShader  = compile ps_2_0 RenderScenePS( false, true, true, false, false, false, false, false );
    }
}

technique RenderSceneToShadowbufferDepth
{
	// actually the same as RenderSceneToShadowbufferIdDepth
    pass P0
    {          
		ALPHABLENDENABLE = false;
		ALPHATESTENABLE = false;
        VertexShader = compile vs_2_0 RenderSceneVS( true, false );
        PixelShader  = compile ps_2_0 RenderScenePS( false, true, true, false, false, false, false, false );
    }
}


